from os import getenv
from time import time
from typing import Any, Callable, Dict, Optional, Union

import jwt
from pydantic import UUID4
from requests import get, post

from llm_monitor.constants.routes import Routes
from llm_monitor.schema.transaction import TransactionRecordBatch
from llm_monitor.utils.request import HttpHeaders, make_request


class ApiClient:
    def __init__(self, project_name: str):
        self.project_id = None
        self.api_url = self.get_api_url()
        if self.healthcheck():
            self.token = self.get_token()
            try:
                self.project_id = self.get_project_id_by_name(project_name)
            except Exception as e:
                if "not found" in str(e):
                    self.create_project(project_name)
                    print(f"🚀 Creating new project... project {project_name} created!")
                else:
                    raise e

    def get_api_url(self) -> str:
        console_url = getenv("GALILEO_CONSOLE_URL")
        if console_url is None:
            raise Exception("GALILEO_CONSOLE_URL must be set")
        if any(map(console_url.__contains__, ["localhost", "127.0.0.1"])):
            api_url = "http://localhost:8088"
        else:
            api_url = console_url.replace("console", "api")
        return api_url

    def get_token(self) -> str:
        username = getenv("GALILEO_USERNAME")
        password = getenv("GALILEO_PASSWORD")
        if username is None or password is None:
            raise Exception("GALILEO_USERNAME and GALILEO_PASSWORD must be set")
        return self.username_login(username, password).get("access_token", "")

    def healthcheck(self) -> bool:
        make_request(get, base_url=self.base_url, endpoint=Routes.healthcheck)
        return True

    def username_login(self, username: str, password: str) -> Dict[str, str]:
        return make_request(
            post,
            base_url=self.base_url,
            endpoint=Routes.login,
            data={
                "username": username,
                "password": password,
                "auth_method": "email",
            },
        )

    @property
    def base_url(self) -> str:
        return self.api_url

    @property
    def auth_header(self) -> Dict[str, str]:
        return {"Authorization": f"Bearer {self.token}"}

    def _make_request(
        self,
        request_method: Callable,
        endpoint: str,
        body: Optional[Dict] = None,
        data: Optional[Dict] = None,
        files: Optional[Dict] = None,
        params: Optional[Dict] = None,
        timeout: Union[int, None] = None,
        json_request_only: bool = False,
    ) -> Any:
        # Check to see if our token is expired before making a request
        # and refresh token if it's expired
        if endpoint is not Routes.login and self.token:
            claims = jwt.decode(self.token, options={"verify_signature": False})
            if claims.get("exp", 0) < time():
                self.token = self.get_token()

        if json_request_only:
            content_headers = HttpHeaders.accept_json()
        else:
            content_headers = HttpHeaders.json()
        headers = {**self.auth_header, **content_headers}
        return make_request(
            request_method=request_method,
            base_url=self.base_url,
            endpoint=endpoint,
            body=body,
            data=data,
            files=files,
            params=params,
            headers=headers,
            timeout=timeout,
        )

    def ingest_batch(self, transaction_batch: TransactionRecordBatch) -> Dict[str, str]:
        return self._make_request(
            post,
            endpoint=Routes.ingest.format(project_id=self.project_id),
            body=transaction_batch.dict(),
        )

    def get_project_id_by_name(self, project_name: str) -> UUID4:
        projects = self._make_request(
            get,
            endpoint=Routes.projects,
            params={"project_name": project_name, "type": "llm_monitor"},
        )
        if len(projects) < 1:
            raise Exception(f"Galileo project {project_name} not found")
        return projects[0]["id"]

    def create_project(self, project_name: str) -> Dict[str, str]:
        return self._make_request(
            post,
            endpoint=Routes.projects,
            body={"name": project_name, "type": "llm_monitor"},
        )



import time
from datetime import datetime
from typing import Any, Dict, List, Optional
from uuid import uuid4

import pytz

from llm_monitor.schema.transaction import TransactionRecord, TransactionRecordType
from llm_monitor.utils.aggregator import (
    add_records_to_batch,
    initialize_api_client,
    start_aggregator_job,
)


class LLMMonitor:
    timers: Dict[str, Dict[str, float]] = {}
    records: Dict[str, TransactionRecord] = {}
    version: str

    def __init__(
        self, project_name: str, version: str, *args: Any, **kwargs: Any
    ) -> None:
        """Initializes LLM Monitor

        Parameters
        ----------
        project_name : str
            The name of the project to log to
        version : str
            A version identifier for this system so logs can be attributed
            to a specific configuration
        """
        self.version = version
        initialize_api_client(project_name=project_name)
        start_aggregator_job()

    def log_prompt(
        self,
        prompt: str,
        model: str,
        temperature: Optional[float] = None,
        chain_id: Optional[str] = None,
    ) -> str:
        """Logs the beginning of a LLM request

        Parameters
        ----------
        prompt : str
            Prompt text as a string
        model : str
            Name of the model being prompted
        temperature : Optional[float]
            Temperature setting being passed to LLM
        chain_id : Optional[str]
            ID of parent if there is one, e.g. chain

        Returns
        -------
        str
            ID of the trace being initiated
        """
        node_id = str(uuid4())

        self.timers[node_id] = {}
        self.timers[node_id]["start"] = time.perf_counter()

        self.records[node_id] = TransactionRecord(
            input_text=prompt,
            model=model,
            temperature=temperature,
            node_id=node_id,
            chain_id=chain_id,
            created_at=datetime.now(tz=pytz.utc).isoformat(),
            node_type=TransactionRecordType.llm.value,
            version=self.version,
        )
        return node_id

    def log_completion(
        self,
        node_id: str,
        output_text: str,
        num_input_tokens: int,
        num_output_tokens: int,
        num_total_tokens: int,
        finish_reason: Optional[str] = None,
        status_code: Optional[int] = None,
        user_metadata: Optional[Dict[str, Any]] = None,
        tags: Optional[List[str]] = None,
    ) -> None:
        """Logs the completion of the LLM request

        Parameters
        ----------
        node_id : str
            ID of the trace started with log_prompt()
        output_text : str
            Completion text from the LLM response
        num_input_tokens : int
            Number of input tokens
        num_output_tokens : int
            Number of output tokens
        num_total_tokens : int
            Total number of tokens
        finish_reason : Optional[str]
            Finish reason from the LLM
        status_code : Optional[int]
            Status code of the API call to the LLM
        user_metadta : Optional[Dict[str, Any]]
            User-defined metadata as key-value pairs
        tags : Optional[List[str]]
            User-defined tags as a list of strings
        """
        self.timers[node_id]["stop"] = time.perf_counter()
        latency_ms = round(
            (self.timers[node_id]["stop"] - self.timers[node_id]["start"]) * 1000
        )
        del self.timers[node_id]

        model_dict = self.records[node_id].model_dump()
        model_dict.update(
            output_text=output_text,
            num_input_tokens=num_input_tokens,
            num_output_tokens=num_output_tokens,
            num_total_tokens=num_total_tokens,
            finish_reason=finish_reason,
            latency_ms=latency_ms,
            status_code=status_code or 200,
            user_metadata=user_metadata,
            tags=tags,
        )

        add_records_to_batch([TransactionRecord(**model_dict)])
        del self.records[node_id]

    def log_error(
        self, node_id: str, error_message: str, status_code: Optional[int]
    ) -> None:
        """Logs an error from an LLM caal

        Parameters
        ----------
        node_id : str
            ID of the trace started with log_prompt()
        error_message : str
            Error message returned from the LLM
        status_code : Optional[int]
            Status code of the API request to the LLM
        """
        self.timers[node_id]["stop"] = time.perf_counter()
        latency_ms = round(
            (self.timers[node_id]["stop"] - self.timers[node_id]["start"]) * 1000
        )
        del self.timers[node_id]

        model_dict = self.records[node_id].model_dump()
        model_dict.update(
            output_text=f"ERROR: {error_message}",
            num_input_tokens=0,
            num_output_tokens=0,
            num_total_tokens=0,
            latency_ms=latency_ms,
            status_code=status_code or 500,
        )

        add_records_to_batch([TransactionRecord(**model_dict)])
        del self.records[node_id]



from enum import Enum
from http.client import HTTPException
from typing import Any, Callable, Dict, Optional, Union
from urllib.parse import urljoin

import backoff
from requests import RequestException, Response


def _validate_response(response: Response) -> None:
    if not response.ok:
        msg = (
            "Something didn't go quite right. The api returned a non-ok status "
            f"code {response.status_code} with output: {response.text}"
        )
        # TODO: Better error handling.
        raise HTTPException(msg)


@backoff.on_exception(backoff.expo, RequestException, max_tries=4, jitter=None)
def make_request(
    request_method: Callable,
    base_url: str,
    endpoint: str,
    body: Optional[Dict] = None,
    data: Optional[Dict] = None,
    files: Optional[Dict] = None,
    params: Optional[Dict] = None,
    headers: Optional[Dict[str, str]] = None,
    timeout: Union[int, None] = None,
) -> Any:
    response = request_method(
        urljoin(base_url, endpoint),
        json=body,
        data=data,
        files=files,
        params=params,
        headers=headers or {},
        timeout=timeout,
    )
    _validate_response(response)
    return response.json()


class HttpHeaders(str, Enum):
    accept = "accept"
    content_type = "Content-Type"
    application_json = "application/json"

    @staticmethod
    def accept_json() -> Dict[str, str]:
        return {HttpHeaders.accept: HttpHeaders.application_json}

    @staticmethod
    def content_type_json() -> Dict[str, str]:
        return {HttpHeaders.content_type: HttpHeaders.application_json}

    @staticmethod
    def json() -> Dict[str, str]:
        return {**HttpHeaders.accept_json(), **HttpHeaders.content_type_json()}



import atexit
import signal
import sys
import threading
from os import getenv
from typing import Any, Callable, List, Optional

from llm_monitor.schema.transaction import TransactionRecord, TransactionRecordBatch
from llm_monitor.utils.api_client import ApiClient

# static
LOOP_INTERVAL = 1
BATCH_SEND_INTERVAL = 10
MAX_BATCH_SIZE = 25
SEND_THRESHOLD = 5
WINDOWS_LIMIT = 5

# variable
BATCH = []
ELAPSED_WINDOWS = 0

IS_STREAMLIT = getenv("IS_STREAMLIT", False)

client: ApiClient


def initialize_api_client(project_name: str) -> None:
    global client
    client = ApiClient(project_name=project_name)


# This method is called by handlers/wrappers to add a new
# record to the send queue
def add_records_to_batch(records: List[TransactionRecord]) -> None:
    global BATCH
    BATCH.extend(records)
    # If we've hit our max BATCH size, don't wait for the next interval
    if len(BATCH) == MAX_BATCH_SIZE:
        _send_batch()


def _send_batch(force_send: bool = False) -> None:
    global BATCH
    global ELAPSED_WINDOWS

    if len(BATCH) > 0:
        if (
            not force_send
            and (len(BATCH) < SEND_THRESHOLD)
            and (ELAPSED_WINDOWS < WINDOWS_LIMIT)
        ):
            ELAPSED_WINDOWS = ELAPSED_WINDOWS + 1
        else:
            try:
                transaction_batch = TransactionRecordBatch(records=BATCH)
                client.ingest_batch(transaction_batch)
                BATCH = []
                ELAPSED_WINDOWS = 0
            except Exception as e:
                print(f"Caught exception in aggregator thread: {e}")


class AggregatorJob(threading.Thread):
    def __init__(self, execute: Callable, *args: Any, **kwargs: Any) -> None:
        threading.Thread.__init__(self)
        self.daemon = True
        self.stopped = threading.Event()
        self.execute = execute
        self.args = args
        self.kwargs = kwargs

        # Send whatever is left in the queue before exiting
        atexit.register(_send_batch, force_send=True)

    def stop(self) -> None:
        self.stopped.set()
        self.join()

    def run(self) -> None:
        elapsed_time = 0
        while not self.stopped.wait(LOOP_INTERVAL):
            if elapsed_time >= BATCH_SEND_INTERVAL:
                self.execute(*self.args, **self.kwargs)
                elapsed_time = 0
            else:
                elapsed_time = elapsed_time + LOOP_INTERVAL


aggregator_job: Optional[AggregatorJob] = None


def _signal_handler(signum: Any, frame: Any) -> None:
    global aggregator_job
    if aggregator_job:
        aggregator_job.stop()
    sys.exit()


def start_aggregator_job() -> None:
    global aggregator_job
    # Prevent duplicate daemon threads if the user initializes
    # the handler multiple times which results in duplicate records
    if aggregator_job is not None:
        return
    if not IS_STREAMLIT:
        # Add signal handlers to make sure we terminate the send_batch job thread
        # Streamlit apps run in a sub-process, so we can't set these
        signal.signal(signal.SIGTERM, _signal_handler)
        signal.signal(signal.SIGINT, _signal_handler)

    # Create the send_batch timed job and start it
    aggregator_job = AggregatorJob(execute=_send_batch)
    aggregator_job.start()



import time
from datetime import datetime
from http import HTTPStatus
from json import dumps
from typing import Any, Dict, List, Optional, Sequence, Union
from uuid import UUID

import pytz
import tiktoken
from langchain.callbacks.base import BaseCallbackHandler
from langchain.prompts.chat import ChatPromptValue
from langchain.schema.agent import AgentAction, AgentFinish
from langchain.schema.messages import BaseMessage
from langchain.schema.messages import _message_to_dict as message_to_dict
from langchain.schema.output import LLMResult
from langchain_core.documents import Document
from pydantic import BaseModel
from pydantic.v1 import BaseModel as BaseModelV1

from llm_monitor.schema.transaction import TransactionRecord, TransactionRecordType
from llm_monitor.utils.aggregator import (
    add_records_to_batch,
    initialize_api_client,
    start_aggregator_job,
)

Stringable = (str, int, float)

# The id field of the serialized response contains a list of identifiers
# that represent the structure of that call like
# ['langchain', 'chat_models', 'openai', 'ChatOpenAI'].
# We want to store the specific constructor being ChatOpenAI
IDX_OF_CONSTRUCTOR = 3


class MonitorHandler(BaseCallbackHandler):
    timers: Dict[str, Dict[str, float]] = {}
    records: Dict[str, TransactionRecord] = {}
    version: str

    def __init__(
        self, project_name: str, version: str, *args: Any, **kwargs: Any
    ) -> None:
        """LangChain callbackbander for LLM Monitoring

        Parameters
        ----------
        project_name : str
            Name of the project to log to
        version : str
            A version identifier for this system so logs can be attributed
            to a specific configuration
        """
        self.version = version
        initialize_api_client(project_name=project_name)
        start_aggregator_job()
        super().__init__(*args, **kwargs)

    def _start_new_node(
        self, run_id: UUID, parent_run_id: Optional[UUID]
    ) -> tuple[str, Optional[str], Optional[str]]:
        node_id = str(run_id)
        chain_id = str(parent_run_id) if parent_run_id else None
        if chain_id:
            # This check ensures we're actually logging the parent chain
            if self.records.get(chain_id):
                self.records[chain_id].has_children = True
                chain_root_id = self.records[chain_id].chain_root_id
            else:
                # We're not logging the parent chain, so this is the root
                chain_root_id = node_id
        else:
            # This node is the root if it doesn't have a parent
            chain_root_id = node_id

        self.timers[node_id] = {}
        self.timers[node_id]["start"] = time.perf_counter()

        return node_id, chain_root_id, chain_id

    def _end_node(self, run_id: UUID) -> tuple[str, int]:
        node_id = str(run_id)

        self.timers[node_id]["stop"] = time.perf_counter()
        latency_ms = round(
            (self.timers[node_id]["stop"] - self.timers[node_id]["start"]) * 1000
        )
        del self.timers[node_id]

        return node_id, latency_ms

    def _finalize_node(self, record: TransactionRecord) -> None:
        self.records[record.node_id] = record
        batch_records: List = []
        # If this record is closing out a root chain, then add all
        # records with that chain_root_id to the batch
        if record.node_id == record.chain_root_id:
            for k, v in self.records.copy().items():
                if v.chain_root_id == record.chain_root_id:
                    batch_records.append(v)
                    del self.records[k]

            add_records_to_batch(batch_records)

    def on_llm_start(
        self,
        serialized: Dict[str, Any],
        prompts: List[str],
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        """Run when LLM starts running."""
        node_id, chain_root_id, chain_id = self._start_new_node(run_id, parent_run_id)
        input_text = prompts[0]
        tags = kwargs.get("tags")
        metadata = kwargs.get("metadata")
        constructor = serialized["id"][IDX_OF_CONSTRUCTOR]
        model = kwargs["invocation_params"]["model_name"]
        temperature = kwargs["invocation_params"].get("temperature")
        self.records[node_id] = TransactionRecord(
            node_id=node_id,
            chain_id=chain_id,
            chain_root_id=chain_root_id,
            input_text=input_text,
            model=model,
            created_at=datetime.now(tz=pytz.utc).isoformat(),
            temperature=temperature,
            tags=tags,
            user_metadata=metadata,
            constructor=constructor,
            node_type=TransactionRecordType.llm.value,
            version=self.version,
        )

    def on_chat_model_start(
        self,
        serialized: Dict[str, Any],
        messages: List[List[BaseMessage]],
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        """Run when Chat Model starts running."""
        node_id, chain_root_id, chain_id = self._start_new_node(run_id, parent_run_id)
        chat_messages = ChatPromptValue(messages=messages[0])
        tags = kwargs.get("tags")
        metadata = kwargs.get("metadata")
        constructor = serialized["id"][IDX_OF_CONSTRUCTOR]
        model = (
            kwargs["invocation_params"]["model"] or kwargs["invocation_params"]["_type"]
        )
        temperature = kwargs["invocation_params"].get("temperature")
        self.records[node_id] = TransactionRecord(
            node_id=node_id,
            chain_id=chain_id,
            chain_root_id=chain_root_id,
            input_text=chat_messages.to_string(),
            model=model,
            created_at=datetime.now(tz=pytz.utc).isoformat(),
            temperature=temperature,
            tags=tags,
            user_metadata=metadata,
            constructor=constructor,
            node_type=TransactionRecordType.chat.value,
            version=self.version,
        )

    def on_llm_end(self, response: LLMResult, run_id: UUID, **kwargs: Any) -> Any:
        """Run when LLM ends running."""
        node_id, latency_ms = self._end_node(run_id)

        generation = response.flatten()[0].generations[0][0]
        output_text = generation.text

        if response.llm_output:
            usage = response.llm_output.get("token_usage", {})
            num_input_tokens = usage.get("prompt_tokens", None)
            num_output_tokens = usage.get("completion_tokens", None)
            num_total_tokens = usage.get("total_tokens", None)
        else:
            try:
                # This is because streaming requests don't provide `llm_output`
                # This only works for OpenAI models (or "" because typing)
                encoding = tiktoken.encoding_for_model(
                    self.records[node_id].model or ""
                )
                num_input_tokens = len(
                    encoding.encode(self.records[node_id].input_text)
                )
                num_output_tokens = len(encoding.encode(output_text))
                num_total_tokens = num_input_tokens + num_output_tokens
            except KeyError:
                num_input_tokens = 0
                num_output_tokens = 0
                num_total_tokens = 0

        if generation.generation_info:
            finish_reason = generation.generation_info.get("finish_reason", "")

        model_dict = self.records[node_id].model_dump()
        model_dict.update(
            output_text=output_text,
            num_input_tokens=num_input_tokens,
            num_output_tokens=num_output_tokens,
            num_total_tokens=num_total_tokens,
            finish_reason=finish_reason,
            latency_ms=latency_ms,
            status_code=HTTPStatus.OK,
        )

        self._finalize_node(TransactionRecord(**model_dict))

    def on_llm_error(self, error: BaseException, run_id: UUID, **kwargs: Any) -> Any:
        """Run when LLM errors."""
        node_id, latency_ms = self._end_node(run_id)

        model_dict = self.records[node_id].model_dump()
        model_dict.update(
            output_text=f"ERROR: {error}",
            num_input_tokens=0,
            num_output_tokens=0,
            num_total_tokens=0,
            latency_ms=latency_ms,
            status_code=getattr(error, "http_status", HTTPStatus.INTERNAL_SERVER_ERROR),
        )

        self._finalize_node(TransactionRecord(**model_dict))

    def on_chain_start(
        self,
        serialized: Dict[str, Any],
        inputs: Dict[str, Any],
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        **kwargs: Any,
    ) -> Any:
        """Run when chain starts running."""
        node_id, chain_root_id, chain_id = self._start_new_node(run_id, parent_run_id)
        tags = kwargs.get("tags")
        metadata = kwargs.get("metadata")
        constructor = serialized["id"][IDX_OF_CONSTRUCTOR]

        if isinstance(inputs, Stringable):
            node_input = {"input": inputs}
        elif isinstance(inputs, BaseMessage):
            node_input = message_to_dict(inputs)
        elif isinstance(inputs, dict):
            # We only support string inputs since those are the likely slots.
            node_input = inputs = {
                key: value
                for key, value in inputs.items()
                if value and isinstance(value, Stringable)
            }
        elif isinstance(inputs, list) and all(isinstance(v, Document) for v in inputs):
            node_input = {
                str(index): value.page_content for index, value in enumerate(inputs)
            }
        else:
            node_input = dict()

        self.records[node_id] = TransactionRecord(
            node_id=node_id,
            chain_id=chain_id,
            chain_root_id=chain_root_id,
            input_text=dumps(
                node_input,
                default=MonitorHandler.json_serializer,
            ),
            created_at=datetime.now(tz=pytz.utc).isoformat(),
            tags=tags,
            user_metadata=metadata,
            node_type=TransactionRecordType.chain.value,
            constructor=constructor,
            version=self.version,
        )

    def on_chain_end(
        self,
        outputs: Union[str, Dict[str, Any]],
        run_id: UUID,
        **kwargs: Any,
    ) -> Any:
        """Run when chain ends running."""
        node_id, latency_ms = self._end_node(run_id)

        if isinstance(outputs, str):
            node_output = {"output": outputs}
        elif isinstance(outputs, dict):
            node_output = outputs
        elif isinstance(outputs, (AgentFinish, AgentAction)):
            node_output = outputs.dict()
        else:
            node_output = dict()

        model_dict = self.records[node_id].model_dump()
        model_dict.update(
            output_text=dumps(
                node_output,
                default=MonitorHandler.json_serializer,
            ),
            finish_reason="chain_end",
            latency_ms=latency_ms,
            status_code=HTTPStatus.OK,
        )

        self._finalize_node(TransactionRecord(**model_dict))

    def on_chain_error(self, error: BaseException, run_id: UUID, **kwargs: Any) -> Any:
        """Run when chain errors."""
        node_id, latency_ms = self._end_node(run_id)

        model_dict = self.records[node_id].model_dump()
        model_dict.update(
            output_text=f"ERROR: {error}",
            latency_ms=latency_ms,
            status_code=getattr(error, "http_status", HTTPStatus.INTERNAL_SERVER_ERROR),
        )

        self._finalize_node(TransactionRecord(**model_dict))

    def on_agent_finish(
        self,
        finish: AgentFinish,
        *,
        run_id: UUID,
        **kwargs: Any,
    ) -> Any:
        """Run on agent end."""
        node_id = str(run_id)
        model_dict = self.records[node_id].model_dump()
        model_dict.update(
            node_type=TransactionRecordType.agent.value,
        )

        self.records[node_id] = TransactionRecord(**model_dict)

    def on_tool_start(
        self,
        serialized: Dict[str, Any],
        input_str: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Any:
        """Run when tool starts running."""
        node_id, chain_root_id, chain_id = self._start_new_node(run_id, parent_run_id)
        tags = kwargs.get("tags")
        metadata = kwargs.get("metadata")
        constructor = serialized["name"]

        self.records[node_id] = TransactionRecord(
            node_id=node_id,
            chain_id=chain_id,
            chain_root_id=chain_root_id,
            input_text=input_str,
            created_at=datetime.now(tz=pytz.utc).isoformat(),
            tags=tags,
            user_metadata=metadata,
            node_type=TransactionRecordType.tool.value,
            constructor=constructor,
            version=self.version,
        )

    def on_tool_end(
        self,
        output: str,
        *,
        run_id: UUID,
        **kwargs: Any,
    ) -> Any:
        """Run when tool ends running."""
        node_id, latency_ms = self._end_node(run_id)

        model_dict = self.records[node_id].model_dump()
        model_dict.update(
            output_text=output,
            latency_ms=latency_ms,
            status_code=HTTPStatus.OK,
        )

        self._finalize_node(TransactionRecord(**model_dict))

    def on_tool_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        **kwargs: Any,
    ) -> Any:
        """Run when tool errors."""
        node_id, latency_ms = self._end_node(run_id)

        model_dict = self.records[node_id].model_dump()
        model_dict.update(
            output_text=f"ERROR: {error}",
            latency_ms=latency_ms,
            status_code=getattr(error, "http_status", HTTPStatus.INTERNAL_SERVER_ERROR),
        )

    def on_retriever_start(
        self,
        serialized: Dict[str, Any],
        query: str,
        *,
        run_id: UUID,
        parent_run_id: Optional[UUID] = None,
        tags: Optional[List[str]] = None,
        metadata: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        """Run on retriever start."""
        node_id, chain_root_id, chain_id = self._start_new_node(run_id, parent_run_id)
        constructor = serialized["id"][-1]

        self.records[node_id] = TransactionRecord(
            node_id=node_id,
            chain_id=chain_id,
            chain_root_id=chain_root_id,
            input_text=str(query),
            created_at=datetime.now(tz=pytz.utc).isoformat(),
            tags=tags,
            user_metadata=metadata,
            node_type=TransactionRecordType.retriever.value,
            constructor=constructor,
            version=self.version,
        )

    def on_retriever_end(
        self,
        documents: Sequence[Document],
        *,
        run_id: UUID,
        **kwargs: Any,
    ) -> None:
        """Run on retriever end."""
        node_id, latency_ms = self._end_node(run_id)

        model_dict = self.records[node_id].model_dump()

        docs = dumps(
            [
                {"page_content": document.page_content, "metadata": document.metadata}
                for document in documents
            ]
        )

        model_dict.update(
            output_text=str(docs),
            finish_reason="retriever_end",
            latency_ms=latency_ms,
            status_code=HTTPStatus.OK,
        )

        record = TransactionRecord(**model_dict)
        self._finalize_node(record)

    def on_retriever_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        **kwargs: Any,
    ) -> None:
        """Run on retriever error."""
        node_id, latency_ms = self._end_node(run_id)

        model_dict = self.records[node_id].model_dump()

        model_dict.update(
            output_text=f"ERROR: {str(error)}",
            latency_ms=latency_ms,
            status_code=getattr(error, "http_status", HTTPStatus.INTERNAL_SERVER_ERROR),
        )

    @staticmethod
    def json_serializer(obj: Any) -> Union[str, Dict[Any, Any]]:
        """
        For serializing objects that cannot be serialized by default with json.dumps.

        Checks for certain methods to convert object to dict.
        """
        if isinstance(obj, BaseModel):
            return obj.model_dump()
        if isinstance(obj, BaseModelV1):
            return obj.dict()
        return str(type(obj))



from enum import Enum
from typing import Any, Dict, List, Optional

from pydantic import BaseModel


class TransactionRecordType(str, Enum):
    llm = "llm"
    chat = "chat"
    chain = "chain"
    tool = "tool"
    agent = "agent"
    retriever = "retriever"


class TransactionRecord(BaseModel):
    latency_ms: Optional[int] = None
    status_code: Optional[int] = None
    input_text: str
    output_text: Optional[str] = None
    model: Optional[str] = None
    num_input_tokens: Optional[int] = None
    num_output_tokens: Optional[int] = None
    num_total_tokens: Optional[int] = None
    finish_reason: Optional[str] = None
    node_id: str
    chain_id: Optional[str] = None
    chain_root_id: Optional[str] = None
    output_logprobs: Optional[Dict] = None
    created_at: str
    tags: Optional[List[str]] = None
    user_metadata: Optional[Dict[str, Any]] = None
    temperature: Optional[float] = None
    constructor: Optional[str] = None
    node_type: TransactionRecordType
    has_children: bool = False
    version: str


class TransactionRecordBatch(BaseModel):
    records: List[TransactionRecord]
